/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.toAntlr;


import java.io.StringReader;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import com.unitesk.atp.bnf.BNFEmptyVisitor;
import com.unitesk.atp.bnf.BNFWalker;
import com.unitesk.atp.bnf.BNFModel.*;

import com.unitesk.atp.bnf.BNF;
import com.unitesk.atp.messages.*;

/*
 * Rule:
 *   antlrDeclCode
 *   antlrInitCode
 *   returnToken
 *   returnNode
 *   returnList
 *   updateList
 *   treedlAbstractNode
 *   treedlBaseNode
 *   treedlSkip
 *
 * Term:
 *   antlrStartActions
 *   enumElementId
 * 
 * Factor (PRE):
 *   antlrLookahead
 *   antlrOptions
 *   antlrPredicate
 *   treedlChild
 *   treedlAttr
 *   treedlAttrType
 *   treedlAttrInitValue
 *   treedlModifier
 *   treedlEnum
 *   antlrListElem
 * 
 * Factor (POST):
 *   antlrActions
 *   treedlCreateNode
 *   treedlCreateNodeParameters
 *   [[antlrResultVar]]
 *   antlrAlternativeResult
 *   antlrTokenTextVar
 * 
 * Lexeme_Factor:
 *   antlrId
 *   antlrVar
 *   antlrResultToken
 * 
 * Name_Factor:
 *   antlrResult
 *
 * Expression_Factor:
 *   treedlEnumElems
 */
public class AntlrAuxilsSetVisitor extends BNFEmptyVisitor
{
    public static BNFWalker createWalker( Map p_token_names
                                        , String p_list_rule_postfix
                                        , String p_list_tail_rule_postfix
                                        , String token_class
                                        , String lxm_prefix
                                        , MessageBox p_mbox
                                        )
    {
        AntlrAuxilsSetVisitor v
            = new AntlrAuxilsSetVisitor
                ( p_token_names
                , p_list_rule_postfix
                , p_list_tail_rule_postfix
                , token_class
                , lxm_prefix
                , p_mbox
                );
        BNFWalker w = new BNFWalker( v, v.new AntlrAuxilsSetPostVisitor() );
        v.walker = w;
        return w;
    }

    private Map token_names;
    private String list_rule_postfix;
    private String list_tail_rule_postfix;
    private String tokenClass;
    private String lxm_prefix;
    private MessageBox mbox;

    private AntlrAuxilsSetVisitor( Map p_token_names
                                 , String p_list_rule_postfix
                                 , String p_list_tail_rule_postfix
                                 , String p_token_class
                                 , String p_lxm_prefix
                                 , MessageBox p_mbox
                                 )
    {
        token_names = p_token_names;
        list_rule_postfix = p_list_rule_postfix;
        list_tail_rule_postfix = p_list_tail_rule_postfix;
        tokenClass = p_token_class;
        lxm_prefix = p_lxm_prefix;
        mbox = p_mbox;
    }

    private BNFWalker walker;

    public void visitLexeme_Factor( Lexeme_Factor node )
    {
        String s
            = node.isText()
            ? AntlrUtils.toAntlrId( node.getText(), token_names )
            : AntlrUtils.toAntlrId( node.getLexemeClass() );
        node.addAttribute( "antlrId", s );

        if( sub_processing ) { return; }

        node.addAttribute( "antlrVar", createAntlrVarId( lxm_prefix ) );

        if( curEnumElem != null )
        {
            if( curEnumElem.length() > 0 ) { curEnumElem += "_"; }
            curEnumElem += s;
        }

        if( returnToken )
        {
            node.addAttribute( "antlrResultToken", "" );
        }

        processFactorDocomment( node );
    }

    public void visitName_Factor( Name_Factor node )
    {
        if( sub_processing ) { return; }

        processFactorDocomment( node );

        if(  !node.hasAttribute("treedlChild")
          && !node.hasAttribute("treedlAttr")
          )
        {
            if(  node.getLabel() != null )
            {
                String id = TreedlUtils.toAttributeId( node.getLabel().getValue() );
                node.addAttribute( "treedlChild", id );
                node.addAttribute( "treedlModifier", modifiers.peek() );
            } else {
                if(  !uniqueUpLevelFactor
                  || node.hasAttribute( "treedlCreateNode" )
                  )
                {
                    String id = TreedlUtils.toAttributeId( node.getName().getValue() );
                    node.addAttribute( "treedlChild", id );
                    node.addAttribute( "treedlModifier", modifiers.peek() );
                } else {
                    String name = node.getName().getValue();
                    node.addAttribute( "antlrResult", "" );
                    if( curRuleName != null && !name.equals( curRuleName ) )
                    {
                        derivedNodes.add( name );
                        Rule r = node.getDefinition();
                        if( r.hasAttribute( "treedlBaseNode" ) )
                        {
System.err.println( "Rule '" + name + "': ambiguous base node ("
                  + r.getAttribute( "treedlBaseNode" ) + ", " + curRuleName + ")"
                  );
                        } else {
                            r.addAttribute( "treedlBaseNode", curRuleName );
                        }
                    }
                }
            }
            //node.addAttribute( "antlrVar", createAntlrVarId( id ) );
        }
    }

    public void visitExpression_Factor( Expression_Factor node )
    {
        if( sub_processing ) { return; }

        char pre = ((String)modifiers.peek()).charAt(0);
        int add = node.getKind();
        String mod = " ";
        if(  pre == '*'
          || add == Expression_Factor.ZERO_LIST
          || add == Expression_Factor.TRANSPOSITION
          || (  pre == '+'
             && add == Expression_Factor.OPTIONAL
             )
          || (  pre == '?'
             && add == Expression_Factor.ONE_LIST
             )
          )
        {
            mod = "*";
        } else if(  pre == '+'
                 || add == Expression_Factor.ONE_LIST
                 ) {
            mod = "+";
        } else if(  pre == '?'
                 || add == Expression_Factor.OPTIONAL
                 ) {
            mod = "?";
        }
        modifiers.push( mod );

        processFactorDocomment( node );
    }

    public void processFactorDocomment( Factor node )
    {
        String jd = node.getOptPreDocomment();
        if( jd != null )
        {
            jd = jd.substring( 3, jd.length() - 2 ).trim();
            String pragma[] = jd.split( "\\s*@\\s*" );
            for( int i = 1; i < pragma.length; i++ )
            {
                if( pragma[i].startsWith("?") )
                {
                    try
                    {
                        String s = pragma[i].substring(1);
                        Expression expr
                            = BNF
                            . createParser( new StringReader( s + ";" ) )
                            . expression();
                        sub_processing = true;
                        expr.accept( walker );
                        sub_processing = false;
                        node.addAttribute( "antlrLookahead", expr );
                    } catch( antlr.RecognitionException e ) {
                        mbox.info( "Error in '?'-pragma '" + pragma[i] + "':" );
                        mbox.info( "      " + e.toString() );
                        mbox.info( "      Pragma ignored" );
                    } catch( antlr.TokenStreamException e ) {
                        mbox.info( "Error in '?'-pragma '" + pragma[i] + "':" );
                        mbox.info( "      " + e.toString() );
                        mbox.info( "      Pragma ignored" );
                    }
                } else if( pragma[i].startsWith("!") ) {
                    String s = pragma[i].substring(1);
                    node.addAttribute( "antlrOptions", s );
                } else if( pragma[i].startsWith("s?") ) {
                    String s = pragma[i].substring(2);
                    node.addAttribute( "antlrPredicate", s );
                } else if( pragma[i].startsWith("$") ) {
                    String init[] = pragma[i].substring(1).trim().split( "\\s*=\\s*" );
                    if( init.length == 1 ) // child
                    {
                        String s = init[0];
                        char last = s.charAt( s.length() - 1 );
                        if( last == '?' || last == '+' || last == '*' )
                        {
                            String id = s.substring( 0, s.length() - 1 );
                            node.addAttribute( "treedlChild", id );
                            //node.addAttribute( "antlrVar", createAntlrVarId( id ) );
                            node.addAttribute( "treedlModifier", "" + last );
                        } else {
                            node.addAttribute( "treedlChild", s );
                            //node.addAttribute( "antlrVar", createAntlrVarId( s ) );
                            node.addAttribute( "treedlModifier", modifiers.peek() );
                        }
                    } else {
                        node.addAttribute( "treedlAttrInitValue", init[1] );
                        String desc[] = init[0].split( ":" );
                        node.addAttribute( "treedlAttr", desc[0] );
                        //node.addAttribute( "antlrVar", createAntlrVarId( desc[0] ) );
                        String t = desc[1];
                        node.addAttribute( "treedlAttrType", t );
                        node.addAttribute( "treedlModifier"
                                         , primitiveTypes.contains( t ) ? " " : modifiers.peek()
                                         );
                    }
                } else if( pragma[i].startsWith("enum") ) {
                    String s = pragma[i].substring(4).trim();
                    node.addAttribute( "treedlAttr", s );
                    //node.addAttribute( "antlrVar", createAntlrVarId( s ) );
                    node.addAttribute( "treedlAttrType", treedlCreateNode + TreedlUtils.toNodeId( s ) );
                    node.addAttribute( "treedlModifier", modifiers.peek() );
                    node.addAttribute( "treedlEnum", "" );
                    enumElems = new Vector();
                } else if( pragma[i].startsWith("elem") ) {
                    String s = pragma[i].substring(4).trim();
                    node.addAttribute( "antlrListElem", "" );
                }
            }
        }

        jd = node.getOptPostDocomment();
        if( jd != null )
        {
            jd = jd.substring( 3, jd.length() - 2 ).trim();
            String pragma[] = jd.split( "\\s*@\\s*" );
            for( int i = 1; i < pragma.length; i++ )
            {
                if( pragma[i].startsWith("{") ) {
                    String s = pragma[i];
                    node.addAttribute( "antlrActions", s );
                } else if(  pragma[i].startsWith("*")
                         || pragma[i].startsWith("|*")
                         ) {
                    if( pragma[i].startsWith("|*") )
                    {
                        node.addAttribute( "antlrAlternativeResult", "" );
                    }
                    String s = pragma[i].substring(pragma[i].indexOf('*') + 1).trim();
                    /*if( s.startsWith( "=" ) )
                    {
                        node.addAttribute( "antlrResultVar", s.substring( 1 ) );
                    } else*/ {
                        int lp = s.indexOf('(');
                        if( lp > 0 )
                        {
                            String params = s.substring( lp + 1, s.length() - 1 ).trim();
                            node.addAttribute( "treedlCreateNodeParameters"
                                             , Arrays.asList( params.split( "\\s*,\\s*" ) )
                                             );
                            s = s.substring( 0, lp );
                        }
                        String name = s.equals( "." ) ? treedlCreateNode : s;
                        node.addAttribute( "treedlCreateNode", name );
                        derivedNodes.add( name );
                    }
                } else if( pragma[i].startsWith("text") ) {
                    String s = pragma[i].substring(4).trim();
                    node.addAttribute( "antlrTokenTextVar", s );
                }
            }
        }
    }

    public void visitTerm( Term node )
    {
        if( sub_processing ) { return; }

        uniqueUpLevelFactor
            =  modifiers.size() == 1
            && node.sizeFactorList() == 1
            ;

        if( enumElems != null )
        {
            curEnumElem = "";
        }

        String jd = node.getFactorList( 0 ).getOptPreDocomment();
        if( jd != null )
        {
            jd = jd.substring( 3, jd.length() - 2 ).trim();
            String pragma[] = jd.split( "\\s*@\\s*" );
            for( int i = 1; i < pragma.length; i++ )
            {
                if( pragma[i].startsWith("{") ) {
                    String s = pragma[i];
                    node.addAttribute( "antlrStartActions", s );
                }
            }
        }
    }

    public void visitRule( Rule node )
    {
        derivedNodes = new HashSet();
        returnToken = false;

        ruleVarCount = new HashMap();
        treedlCreateNode = null;
        String jd = node.getOptDocomment();
        if( jd != null )
        {
            jd = jd.substring( 3, jd.length() - 2 ).trim();
            String pragma[] = jd.split( "\\s*@\\s*" );
            for( int i = 1; i < pragma.length; i++ )
            {
                if( pragma[i].startsWith("declCode") ) {
                    String s = pragma[i].substring(8).trim();
                    node.addAttribute( "antlrDeclCode", s );
                } else if( pragma[i].startsWith("initCode") ) {
                    String s = pragma[i].substring(8).trim();
                    node.addAttribute( "antlrInitCode", s );
                } else if( pragma[i].startsWith("*") ) {
                    treedlCreateNode = pragma[i].substring(1).trim();
                    //node.addAttribute( "treedlCreateNode", treedlCreateNode );
                    node.addAttribute( "returnNode", treedlCreateNode );
                } else if( pragma[i].startsWith("token") ) {
                    node.addAttribute( "returnToken", tokenClass );
                    returnToken = true;
                } else if( pragma[i].startsWith("abstract") ) {
                    node.addAttribute( "treedlAbstractNode", "" );
                } else if( pragma[i].startsWith("base") ) {
                    String s = pragma[i].substring(4).trim();
                    node.addAttribute( "treedlBaseNode", s );
                }
            }
        }
        String name = node.getName().getValue();
/*
        if( treedlCreateNode == null )
        {
            treedlCreateNode = name;
            node.addAttribute( "treedlCreateNode", treedlCreateNode );
        }
*/
        if( name.endsWith( list_rule_postfix ) )
        {
            node.addAttribute( "returnList", "" );
            //isList = true;
        } else if( name.endsWith( list_tail_rule_postfix ) )
        {
            node.addAttribute( "updateList", "" );
            //isList = true;
        } else if( node.hasAttribute("returnNode") )
        {
            node.addAttribute( "treedlSkip", "" );
        } else if(  !node.hasAttribute("returnToken")
//                  && !node.hasAttribute("returnNode")
                 ) {
            treedlCreateNode = name;
            node.addAttribute( "returnNode", treedlCreateNode );
            curRuleName = name;
        }
    }

    public void visitGrammar( Grammar node )
    {
        //grammar = node;
        modifiers.push( " " );
    }

    private boolean sub_processing = false;

    private Stack modifiers = new Stack();
    private List enumElems = null;
    private String curEnumElem = null;
    private String treedlCreateNode = null;

    private Set derivedNodes = null;
    private String curRuleName = null;
    //private Grammar grammar = null;

    private Map ruleVarCount;
    private boolean uniqueUpLevelFactor = false;

    private boolean returnToken = false;

    private Set primitiveTypes = new HashSet();
    {
        primitiveTypes.add( "bool" );
        primitiveTypes.add( "char" );
        primitiveTypes.add( "short" );
        primitiveTypes.add( "int" );
        primitiveTypes.add( "long" );
        primitiveTypes.add( "float" );
        primitiveTypes.add( "double" );
    }

    private String createAntlrVarId( String id )
    {
        Object n = ruleVarCount.get( id );
        if( n == null )
        {
            ruleVarCount.put( id, new Integer( 0 ) );
            return id;
        } else {
            int count = ((Integer)n).intValue() + 1;
            ruleVarCount.put( id, new Integer( count ) );
            return id + "_" + count;
        }
    }

    private class AntlrAuxilsSetPostVisitor extends BNFEmptyVisitor
    {
        public void visitGrammar( Grammar node )
        {
            modifiers.pop();
            //grammar = null;
        }
        public void visitRule( Rule node )
        {
            if(  derivedNodes.size() > 1
              && !node.hasAttribute( "treedlAbstractNode" )
              )
            {
                node.addAttribute( "treedlAbstractNode", "" );
            }
            derivedNodes = null;
            curRuleName = null;
        }
        public void visitTerm( Term node )
        {
            if( sub_processing ) { return; }

            if( curEnumElem != null )
            {
                node.addAttribute( "enumElementId", curEnumElem );
                enumElems.add( curEnumElem );
                curEnumElem = null;
            }
        }
        public void visitExpression_Factor( Expression_Factor node )
        {
            if( sub_processing ) { return; }

            modifiers.pop();
            if( enumElems != null )
            {
                node.addAttribute( "treedlEnumElems", enumElems );
                enumElems = null;
            }
        }
    }
}
