/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.atp.bnf.actions.toAntlr;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tool;
import com.unitesk.atp.tree.tool.Tree;
import com.unitesk.atp.tree.generation.TextGeneratorServer;
import com.unitesk.atp.text.filters.ExceptionHandler;
import com.unitesk.atp.text.filters.WriterTextReceiver;
import com.unitesk.atp.messages.*;

import com.unitesk.atp.bnf.BNFWalker;
import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.auxil.FileUtils;

public class PrintAntlrGrammarAction extends Action implements ExceptionHandler
{
    private File cur_output_file;

    public void process( Exception e )
    {
        mbox.fatal( new IOErrorMessage( cur_output_file ) );
    }

    public void process( Tree tree )
    {
        Grammar grammar = (Grammar)tree.getRoot();
        checkSemantics( tree );
        mbox.status();

        boolean update_mode = getProperty( "update", "true" ).equals( "true" );

        String antlr_output_file_name = getProperty( "antlr.output.file" );
        if( antlr_output_file_name == null )
        {
            mbox.info( "Property 'antlr.output.file' is not defined. Action ignored." );
            return;
        }

        File antlr_output_file = new File( antlr_output_file_name );
        File parentFile = antlr_output_file.getParentFile(); 
        if( parentFile != null ) parentFile.mkdirs();

        String antlr_preamble_name = getProperty( "antlr.preamble.file" );
        File antlr_preamble_file = antlr_preamble_name == null ? null : new File( antlr_preamble_name );

        String token_class = getProperty( "antlr.token.class", "Token" ); // ???
        String model_class = getProperty( "treedl.model.class", "Model" ); // ???

        String var_name_prefix = getProperty( "antlr.var.name.prefix", "v_" );
        String list_var_postfix = getProperty( "antlr.list.var.postfix", "_list" );

        {
            /* Set antlrIds to terminals */
            BNFWalker w = AntlrAuxilsSetVisitor.createWalker
                ( FileUtils.loadMap( this, "antlr.token.names.file", true )
                , getProperty( "antlr.list.rule.postfix", "List" )
                , getProperty( "antlr.list.tail.rule.postfix", "ListTail" )
                , "<" + token_class + ">"
                , getProperty( "antlr.lexeme.label.prefix", "lxm__" )
                , mbox
                );
            grammar.accept( w );
        }

        ListElemTypeHandler.handle( grammar, "<" + token_class + ">" );

        TreedlFieldsCollector.collect( grammar, "<" + token_class + ">" );

        AntlrVarDeclsSetVisitor.process( grammar, model_class, var_name_prefix, list_var_postfix );

        if(  update_mode 
          && pm.getTree().getFile().lastModified() < antlr_output_file.lastModified()
          && (  antlr_preamble_file != null
             && antlr_preamble_file.lastModified() < antlr_output_file.lastModified()
             )
          )
        {
            mbox.info( "File is up to date: " + antlr_output_file );
        } else {
            try
            {
                cur_output_file = antlr_output_file;
                mbox.info( "-> " + antlr_output_file );
                if( antlr_preamble_file != null )
                {
                    FileUtils.copyFile( antlr_preamble_file, antlr_output_file );
                }
                FileWriter file_writer
                    = new FileWriter( antlr_output_file, antlr_preamble_file != null ); // true -- append
                TextGeneratorServer s 
                    = new TextGeneratorServer( new WriterTextReceiver( file_writer, this ) );
                ParserRulesPrinterVisitor v = new ParserRulesPrinterVisitor();

                s.setVariable( "method_name_prefix", getProperty( "antlr.method.name.prefix", "m_" ) );
                s.setVariable( "var_name_prefix", var_name_prefix );
                s.setVariable( "list_var_postfix", list_var_postfix );
                s.setVariable( "model_class", model_class );
                s.setVariable( "token_class", token_class );
                s.setVariable( "start_token", getProperty( "antlr.start.token.var", "__start__token__" ) );
                s.setVariable( "result_var", getProperty( "antlr.rule.result.var", "result__value__" ) );
                s.setVariable( "list_elem_var", getProperty( "antlr.list.elem.var", "list__element__" ) );

                s.setVariable( "node_factory", getProperty( "parser.nodeFactory.var", "nodeFactory" ) );
                s.setVariable( "list_factory", getProperty( "parser.listFactory.var", "listFactory" ) );
                s.setVariable( "set_range", getProperty( "parser.setRange.method", "setRange" ) );

                s.setFunction( "main", v );
                s.setFunction( "declVar", new AntlrVarDeclPrinterVisitor() );
                s.setFunction( "initVar", new AntlrVarInitPrinterVisitor() );

                s.setVariable( "toolName", pm.getName() );
                if( pm instanceof Tool )
                {
                    s.setVariable( "toolVersion", ((Tool)pm).getVersion() );
                }
                s.setVariable( "sourceFileName"
                             ,   pm.getTree().getFile() != null 
                               ? pm.getTree().getFile().getName() 
                               : "" 
                             );
                v.process( grammar );
                file_writer.close();
            }
            catch( IOException e )
            {
                mbox.info( e.getMessage() );
                mbox.fatal( new IOErrorMessage( antlr_output_file ) );
            }
        }


        String treedl_output_file_name = getProperty( "treedl.output.file" );
        if( treedl_output_file_name == null )
        {
            return;
        }

        File treedl_output_file = new File( treedl_output_file_name );
        parentFile = treedl_output_file.getParentFile(); 
        if( parentFile != null ) parentFile.mkdirs();

        String treedl_properties_name = getProperty( "treedl.properties.file" );
        File treedl_properties_file
            = treedl_properties_name == null ? null : new File( treedl_properties_name );

        String treedl_preamble_name = getProperty( "treedl.preamble.file" );
        File treedl_preamble_file = treedl_preamble_name == null ? null : new File( treedl_preamble_name );

        if(  update_mode 
          && pm.getTree().getFile().lastModified() < treedl_output_file.lastModified()
          && (  treedl_preamble_file != null
             && treedl_preamble_file.lastModified() < treedl_output_file.lastModified()
             )
          && (  treedl_properties_file != null
             && treedl_properties_file.lastModified() < treedl_output_file.lastModified()
             )
          )
        {
            mbox.info( "File is up to date: " + treedl_output_file );
        } else {
            try
            {
                if( treedl_properties_file != null )
                {
                    File treedl_properties_output_file = new File( treedl_output_file_name + ".properties" );
                    mbox.info( "-> " + treedl_properties_output_file );
                    FileUtils.copyFile( treedl_properties_file, treedl_properties_output_file );
                }

                cur_output_file = treedl_output_file;
                mbox.info( "-> " + treedl_output_file );
                if( treedl_preamble_file != null )
                {
                    FileUtils.copyFile( treedl_preamble_file, treedl_output_file );
                }
                FileWriter file_writer
                    = new FileWriter( treedl_output_file, treedl_preamble_file != null ); // true -- append
                TextGeneratorServer s 
                    = new TextGeneratorServer( new WriterTextReceiver( file_writer, this ) );
                TreedlNodesPrinterVisitor v = new TreedlNodesPrinterVisitor
                    ( FileUtils.loadMap( this, "treedl.late.attributes.file", false ) );
                s.setFunction( "nodes", v );
                s.setFunction( "enums", new TreedlEnumsPrinterVisitor() );

                s.setVariable( "model_class", model_class );
                s.setVariable( "base_node", getProperty( "treedl.base.node" ) );

                s.setVariable( "toolName", pm.getName() );
                if( pm instanceof Tool )
                {
                    s.setVariable( "toolVersion", ((Tool)pm).getVersion() );
                }
                s.setVariable( "sourceFileName"
                             ,   pm.getTree().getFile() != null 
                               ? pm.getTree().getFile().getName() 
                               : "" 
                             );
                v.process( grammar );
                file_writer.close();
            }
            catch( IOException e )
            {
                mbox.info( e.getMessage() );
                mbox.fatal( new IOErrorMessage( treedl_output_file ) );
            }
        }
    }

    private boolean isSemanticsChecked( Grammar grammar )
    {
        return (  grammar.hasAttribute( "checked" )
               && grammar.getAttribute( "checked" ) instanceof Boolean
               && ((Boolean)grammar.getAttribute( "checked" )).booleanValue()
               );
    }

    private void checkSemantics( Tree tree )
    {
        if( !isSemanticsChecked( (Grammar)tree.getRoot() ) )
        {
            Object checker = getPlugin( getProperty( "checker", "check" ) );
            if( checker == null || !(checker instanceof Action) ) { return; }
            ((Action)checker).process( tree );
        }
    }

    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return com.unitesk.atp.tool.PluginClass.getRevision( "$Revision$" );
    }
}

