/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.atp.bnf;

//==============================================================================

import java.io.File;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.Reader;

import antlr.RecognitionException;
import antlr.TokenStreamBasicFilter;
import antlr.TokenStreamException;
import antlr.TokenStreamRecognitionException;

import com.unitesk.atp.tree.tool.antlr.*;
import com.unitesk.atp.tree.tool.Tool;
import com.unitesk.atp.tree.tool.Tree;
import com.unitesk.atp.bnf.BNFModel.Grammar;
import com.unitesk.atp.messages.*;
import com.unitesk.atp.text.location.Position;

//==============================================================================

/**
 * @author <A href="mailto:demakov@ispras.ru">Alexey Demakov</A>
 * @version $Id$
 */
public abstract class BNFLauncher extends Tool
{
    private static MessageBox main_mbox;
    public BNFLauncher()
    {
        super();
        main_mbox = mbox;
    }

    public static BNF_Parser createParser( Reader r )
    {
        BNF_Lexer lexer = new BNF_Lexer( r );
        lexer.setTokenObjectClass( "com.unitesk.atp.tree.tool.antlr.LinkedToken" );

        LinkFilter lf = new LinkFilter( lexer );
        TokenStreamBasicFilter bf = new TokenStreamBasicFilter( lf );

        bf.discard( BNF_LexerTokenTypes.WS );
        bf.discard( BNF_LexerTokenTypes.COMMENT );
        bf.discard( BNF_LexerTokenTypes.SL_COMMENT );
        bf.discard( BNF_LexerTokenTypes.ML_COMMENT );

        return new BNF_Parser( bf, null, main_mbox );
    }

    public void parse( File file )
    {
        try
        {
            BNF_Lexer lexer = new BNF_Lexer( new FileReader( file ) );
            lexer.setTokenObjectClass( "com.unitesk.atp.tree.tool.antlr.LinkedToken" );

            LinkFilter lf = new LinkFilter( lexer );
            TokenStreamBasicFilter bf = new TokenStreamBasicFilter( lf );

            bf.discard( BNF_LexerTokenTypes.WS );
            bf.discard( BNF_LexerTokenTypes.COMMENT );
            bf.discard( BNF_LexerTokenTypes.SL_COMMENT );
            bf.discard( BNF_LexerTokenTypes.ML_COMMENT );
            
            BNF_Parser parser = new BNF_Parser( bf, file, mbox );

            Grammar g = parser.grammar();
            mbox.status();

            g.setBnfFileName( file );
            g.setFirstToken( lf.getFirstToken() );

            LinkedToken t = lf.getLastToken();
            if( t.getType() == BNF_LexerTokenTypes.EOF )
            {
                // remove EOF
                t = t.getPrevToken();
                if( t != null )
                {
                    t.setNextToken( null );
                }
            }
            g.setLastToken( t );

            Tree tree = new Tree( this );
            tree.setFile( file );
            tree.setRoot( g );
            setTree( tree );
        }
        catch( FileNotFoundException e )
        {
            mbox.fatal( new FileNotFoundMessage( file ) );
        }
        catch( TokenStreamRecognitionException e )
        {
            mbox.fatal( new SyntaxErrorMessage
                            ( file
                            , new Position( e.recog.getLine()
                                          , e.recog.getColumn()
                                          )
                            , e.getMessage()
                            )
                      );
        }
        catch( TokenStreamException e )
        {
            mbox.fatal( new IOErrorMessage( file ) );
        }
        catch( RecognitionException e )
        {
            mbox.fatal( new SyntaxErrorMessage
                            ( file
                            , new Position( e.getLine()
                                          , e.getColumn()
                                          )
                            , e.getMessage()
                            )
                      );
        }
    }
}
