/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

//==============================================================================

package com.unitesk.atp.bnf.actions.testgen;

//==============================================================================

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Vector;

import com.unitesk.atp.tree.Node;
import com.unitesk.atp.tree.tool.Action;
import com.unitesk.atp.tree.tool.Tree;

import com.unitesk.atp.text.location.Position;
import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.auxil.*;
import com.unitesk.atp.bnf.actions.testgen.auxil.*;
import com.unitesk.atp.bnf.actions.testgen.backend.*;
import com.unitesk.atp.bnf.messages.*;
import com.unitesk.atp.messages.*;

//==============================================================================

public class AlternativesPosTestGenerator extends Action
{
    private static String version = "1.1.2";

    private Vector/*Rule*/ order;
    private Rule start_rule;
    private LexemeTable lexeme_table;
    private TestCaseProcessor processor;
    private AllGenerator all_generator = new AllGenerator();
    private ChainGenerator chain_generator = new ChainGenerator();
    private FirstRealizationGenerator first_realization_generator;
    private int first_test_number = 0;
    private int last_test_number = 0;
    private boolean done = false;

    private int point = 0;
    private boolean was_point = false;

    private int getIntProperty( String s, int def )
    {
        String prop = getProperty( s );
        if( prop == null || prop.length() <= 0 ) return def;
        else return Integer.parseInt( prop );
    }

    private void init( Grammar grammar )
    {
        String start_rule_name = getProperty( "start.rule" );
        String test_lexeme_file = getProperty( "lexeme.file" );
        String test_printer = getProperty( "processor" );

        if ( !isSemanticsChecked( grammar ) )
        {
            mbox.fatal( new PropertyErrorMessage( getName()
                                                , "checker"
                                                , "Grammar is not checked"
                                                )
                      );
        }

        if( start_rule_name == null )
        {
            mbox.fatal( new PropertyErrorMessage( getName()
                                                , "start.rule"
                                                , "Start rule is not defined"
                                                )
                      );
        } else {
            start_rule = grammar.getRule( start_rule_name );
            if( start_rule == null )
            {
                mbox.fatal( new PropertyErrorMessage( getName()
                                                    , "start.rule"
                                                    , "No such rule: " + start_rule_name
                                                    )
                          );
            }
        }

        if( test_lexeme_file == null )
        {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "lexeme.file"
                                                , "Lexeme file is not defined"
                                                )
                      );
        } else {
            lexeme_table = new LexemeTable( test_lexeme_file, mbox );
        }

        if( test_printer == null )
        {
            processor = null;
            mbox.warning( new PropertyErrorMessage( getName()
                                                  , "processor"
                                                  , "Test case processor is not defined"
                                                  )
                        );
        } else {
            Object o = getPlugin( test_printer );

            if( o == null )
            {
                mbox.error( new PropertyErrorMessage( getName()
                                                    , "processor"
                                                    , "No such test case processor: "
                                                    + test_printer
                                                    )
                          );
            } else if( !(o instanceof TestCaseProcessor) ) {
                mbox.error( new PropertyErrorMessage( getName()
                                                    , "processor"
                                                    , "Plugin '" + test_printer
                                                    + "' is not a test case processor"
                                                    )
                          );
            } else {
                processor = (TestCaseProcessor)o;
            }
        }

        mbox.status();

        int first_depth = getIntProperty( "first.depth", 2 );
        int all_depth   = getIntProperty( "all.depth", 2 );
        int stop_depth  = getIntProperty( "stop.depth", 0 );
        String fr_file_name = getProperty( "first.realization.file" );

        try // ???
        {
            first_realization_generator
                = new FirstRealizationGenerator( grammar
                                               , fr_file_name
                                               , first_depth
                                               , new BoundMarker()
                                               , lexeme_table
                                               , mbox
                                               );
        } catch( InvalidLexemeValueException e ) {
            mbox.fatal( new ExceptionMessage( e, "initializing plugin '" + getName() + "'" ) );
        } catch( LexemeNotFoundException e ) {
            mbox.fatal( new ExceptionMessage( e, "initializing plugin '" + getName() + "'" ) );
        }
        all_generator.setMaxDepth( all_depth );
        all_generator.setDepth( stop_depth );

        first_test_number = getIntProperty( "first.test", 0 );
        int number = getIntProperty( "test.quantity", 0 );
        if ( number > 0 ) last_test_number = number + first_test_number;
    }

    private boolean isSemanticsChecked( Grammar grammar )
    {
        return (  grammar.hasAttribute( "checked" )
               && grammar.getAttribute( "checked" ) instanceof Boolean
               && ((Boolean)grammar.getAttribute( "checked" )).booleanValue()
               );
    }

    private void checkSemantics( Tree tree )
    {
        if( !isSemanticsChecked( (Grammar)tree.getRoot() ) )
        {
            Object checker = getPlugin( getProperty( "checker", "check" ) );
            if( checker == null || !(checker instanceof Action) ) { return; }
            ((Action)checker).process( tree );
        }
    }

    public void process( Tree tree )
    {
        if( !( tree.getRoot() instanceof Grammar ) )
        {
            mbox.warning
                ( new PlainTextMessage
                    ( "Structure to be processed is not a BNF-grammar. Action ignored" )
                );
            return;
        }
        Grammar grammar = (Grammar)tree.getRoot();

        checkSemantics( tree );
        init( grammar );
        setOrder( grammar );
        mbox.status();

        printOrder();

        // 
        generateAllTests();

        if ( processor != null )
        {
            processor.close();
        }
        mbox.status();
    }

    private void generateAllTests()
    {
        mbox.info( "Generating tests" );
        for( int i = 0; i < order.size(); i++ )
        {
            String id = "Step " + (i+1) + " of " + order.size();
            Rule rule = (Rule)order.elementAt(i);
            String rule_name = rule.getName().getValue();

            // get chain from start rule to current rule
            Vector/*Rule*/ chain = ChainBuilder.getChain( start_rule, rule );

            if( chain == null )
            {
                mbox.warning
                    ( new PlainTextMessage
                        ( id + ". Rule '" + rule_name + "' is unreachable (ignored)" )
                    );
                continue;
            }

            mbox.info( id + ". Generating tests for rule '" + rule_name + "'" );

            // generate tests    
            generateTests( chain );

            if( was_point )
            {
                System.err.println();
                was_point = false;
                point = 0;
            }

            mbox.info( "    Number of tests for rule '" + rule_name
                     + "' = " + test_for_rule_counter
                     );
            test_for_rule_counter = 0;
            if( done ) 
            {
                break;
            }
        }
        mbox.info( "Number of all generated tests = " + test_counter );
    }

    private void generateTests( Vector/*Rule*/ chain )
    {
        all_generator.clear();
        chain_generator.clear();

        generate_chain( chain
                      , 1
                      , new NonTerminal(null, (Rule)chain.firstElement()) 
                      );
    }

    private void setOrder( Grammar grammar )
    {
        String file_name = getProperty( "order.file" );
        order = new Vector();
        BufferedReader in = null;
        if ( file_name != null )
        {
            mbox.info( "Reading rule processing order from file: " + file_name );
            try
            {
                in = new BufferedReader( new FileReader( file_name ) );
            }
            catch( FileNotFoundException e )
            {
                mbox.warning( new FileNotFoundMessage( new File( file_name ) ) );
                return;
            }
        }
        if( in != null )
        {
            int line_num = 1;
            try
            {
                for( String s = null; (s = in.readLine()) != null; line_num++ )
                {
                    s = s.trim();
                    Rule rule = grammar.getRule( s );
                    if( rule != null )
                    {
                        order.addElement( rule );
                    } else {
                        mbox.warning
                            ( new SyntaxErrorMessage( new File( file_name )
                                                    , new Position( line_num, 0 )
                                                    , "No such rule: '" + s + "'"
                                                    )
                            );
                    }
                }
                in.close();
            }
            catch( IOException e )
            {
                mbox.error( new IOErrorMessage( new File( file_name ) ) );
            }
        } else {
            mbox.info( "Generating priority-based rule processing order" );
            Hashtable/*Rule,Vector/Rule/*/ priority = new Hashtable();
            for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
            {
                Rule rule = grammar.getOptRuleList( i );
                priority.put( rule, new Vector());
            }
            
            for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
            {
                Rule  rule = grammar.getOptRuleList( i );
                Vector/*Rule*/ chain = ChainBuilder.getChain( start_rule, rule );
                if( chain == null )
                {
                    mbox.warning
                        ( new PlainTextMessage
                            ( "Rule '" + rule.getName().getValue() + "' is unreachable (ignored)" )
                        );
                    continue;
                }
                if ( chain.size() > 1 )
                {
                    Rule r = (Rule)chain.elementAt( chain.size() - 2 );
                    ((Vector)priority.get( r )).addElement( rule );
                }            
            }
            order.addElement( start_rule );
            iteratePriorityTree( start_rule, priority );
        }
    }

    private void iteratePriorityTree( Rule r, Hashtable/*Rule,Vector/Rule/*/ priority )
    {
        Vector ch = (Vector)priority.get(r);
        for( int i = 0; i < ch.size(); i++ )
        {
            boolean flg = false;
            for ( int j = 0; j < order.size(); j++)
            {
                if (  ((Rule)(order.elementAt(j))).getName().getValue()
                     .equals(
                      ((Rule)(   ch.elementAt(i))).getName().getValue()
                            )
                   )
                { flg = true; }
            }
            if ( !flg )
            {
                order.addElement( (Rule)ch.elementAt(i) );
                iteratePriorityTree( (Rule)ch.elementAt(i), priority );
            }
        }    
    }

    private void printOrder()
    {
        mbox.info( "Order of processing BNF-rules to generate tests:" );
        for ( int i = 0; i < order.size(); i++ )
        {
            mbox.info( "    Rule " + (i+1) + ": "
                     + ((Rule)order.elementAt(i)).getName().getValue()
                     );
        }
    }

    private void sendTestToProsessor( Sentence s )
    {
        test_counter++;
        test_for_rule_counter++;
        if ( test_counter <= first_test_number ) return;
        if ( last_test_number != 0 && test_counter >= last_test_number ) done = true;

        if ( processor != null )
        {
            try
            {
                processor.process( s, "positive-alternatives" );
            } catch ( TestCaseProcessorException e )
            {
                test_for_rule_counter--;
                test_counter--;
                if( was_point )
                {
                    System.err.println();
                    was_point = false;
                }
                mbox.info( "    Number of tests for last processed rule = " 
                         + test_for_rule_counter
                         );
                mbox.info(  "Number of all generated tests == " + test_counter );
                processor.close();
                mbox.fatal( new ExceptionMessage( e, "processing test case" ) );
            }
        }

        point++;
        if( point >= 1000 )
        {
            System.err.print( "." );
            was_point = true;
            point = 0;
        } 
    }
    
    private int test_counter = 0;
    private int test_for_rule_counter = 0;

    /**
     *      .
     *
     * @param chain      ,  
     *   .
     * @param index  ,     .
     */
    private Boolean generate_chain( Vector/*Rule*/ chain, int index, NonTerminal nt )
    {
        if( index < chain.size() )
        {
            /*
             *      . 
             *  ,    . 
             *       -  
             * .
             *     .
             */
            nt.setLabel( false );
            Vector/*Rule*/ ch = chain_generator.setChain( chain );
            int i = chain_generator.setIndex( index );

            Boolean res = (Boolean)nt.realize( chain_generator );

            chain_generator.setChain( ch );
            chain_generator.setIndex( i );
            return res;
        } else {
            /* 
             *    ,    
             * .
             */
            if ( chain_generator.getBegin() == null 
               || chain_generator.getEnd() == null )
            {
                mbox.warning
                    ( new PlainTextMessage
                        ( "Can not generate begin and end for current rule."
                        + " Possible, some first-realization is absent."
                        )
                    );
                return new Boolean( false );
            }

            nt.setLabel( false );
            all_generator.setLevel(-1);
            all_generator.setBegin( chain_generator.getBegin() );
            all_generator.setEnd( chain_generator.getEnd() );

            Boolean res = (Boolean)nt.realize( all_generator );
            return res;
        }
    }

    private Sentence generate_first( Node node )
    {
        return first_realization_generator.generate_first(node);
    }

    //==========================================================================

    private class BoundMarker implements SentenceMarker
    {
        public Sentence mark(Sentence s, Rule r)
        {
            if ( s != null )
            {
                Bound begin = new Bound( r );
                Sentence res = new Sentence( begin );
                res.append(s);
                res.append( begin.getEnd() );
                return res;
            }
            return s;
        }
     }

//==============================================================================

    private class Bound extends Word
    {
        public Bound( Rule p_rule )
        {
            rule = p_rule;
            end = new End();
        }
        
        public Bound( NonTerminal p_rule )
        {
            this(p_rule.getRule());
        }
        
        public void print( PrintWriter out )
        {
            out.print( "<rule name=\"" + getRuleName() + "\">" );
        }
        
        public Word getEnd() { return end; }
        
        public String getRuleName() { return rule.getName().getValue(); }
        
        private Rule rule;
        private End end;
        
        private class End extends Word
        {
            public void print( PrintWriter out ) { out.print("</rule>"); }
        }
    }

//==============================================================================

    private class AllGenerator implements NonTerminalRealizer
    {
        public void setMaxDepth( int p_max_depth )
        {
            max_depth = p_max_depth;
        }
        
        public void setDepth( int p_max_depth )
        {
            depth = p_max_depth;
        }
        
        public NonTerminal setFather( NonTerminal p_father )
        {
            NonTerminal f = father;
            father = p_father;
            return f;
        }
        
        public void setLevel( int l ) { level = l; }
        
        public Object getResult() { return result; }
        
        public void visitNode( Node node ) {}

        public void visitGrammar( Grammar node ) {}
        
        public void visitRule( Rule node )
        {
            if( AlternativesPosTestGenerator.this.done ) return;
            level++;
            Bound beg = new Bound( node );
            if( begin == null ){ begin = new Sentence(); }
            begin.append( beg );
            if( end == null ){ end = new Sentence(); }
            Sentence res = new Sentence( beg.getEnd() );
            res.append( end );
            end = res;

            node.getExpression().accept( this );
            level--;
        }
        
        public void visitExpression( Expression node )
        {
            if( AlternativesPosTestGenerator.this.done ) return;
            for( int i = 0; i < node.sizeTermList(); i++ )
            {
                node.getTermList( i ).accept( this );
            }
        }

        public void visitTerm( Term node )
        {
            if( AlternativesPosTestGenerator.this.done ) return;
            for (int j = 0; j < node.sizeFactorList(); j++ )
            {
                if( AlternativesPosTestGenerator.this.done ) return;
                if( level != 0 && j != 0 ) break; //   "A F F"
            
                Sentence b = begin.makeClone();
                Sentence res = begin;
                for( int i = 0; i < j; i++ )
                {                    
                    Factor factor = node.getFactorList( i );
                    Sentence res_first = AlternativesPosTestGenerator.this.generate_first( factor );
                    if( res_first == null ) { return; }
                    res.append( res_first );
                }
                begin = res;
                Sentence e = end.makeClone();;
                res = new Sentence();
                for( int i = j + 1; i < node.sizeFactorList(); i++ )
                {         
                    Factor factor = node.getFactorList( i );
                    Sentence res_first = AlternativesPosTestGenerator.this.generate_first( factor );
                    if( res_first == null )
                    {
                        begin = b;
                        return;
                    }
                    res.append( res_first );
                }
                res.append(end);
                end = res;
                
                Factor factor = node.getFactorList( j );

                factor.accept( this );

                begin = b;
                end = e;
            }
        }

        public void visitFactor( Factor node ) {}
        
        public void visitName_Factor( Name_Factor node )
        {
            if( AlternativesPosTestGenerator.this.done ) return;
            if(    father == null
                || (    father.getDepth( node.getName().getValue() ) < max_depth
                     && ( depth == 0 || level < depth ) 
                   )
              )
            {
                result = (Boolean)new NonTerminal( father
                                                 , node.getDefinition()
                                                 ).realize( this );
            }
        }

        public void visitLexeme_Factor( Lexeme_Factor node )
        {
            if( AlternativesPosTestGenerator.this.done ) return;
            String l = node.getLexeme();
            for( int i = 0; i < lexeme_table.sizeLexeme( l ); i++ )
            {
                Sentence res = new Sentence();
                res.append( begin );
                res.append( new Terminal( lexeme_table.getLexeme( l, i ) ) );
                res.append( end );
                
                AlternativesPosTestGenerator.this.sendTestToProsessor( res );
            }
            result = new Boolean( true );
        }

        public void visitExpression_Factor( Expression_Factor node )
        {
            if( AlternativesPosTestGenerator.this.done ) return;
            switch( node.getKind() )
            {
            case Expression_Factor.OPTIONAL:
                {
                    Sentence res = new Sentence();
                    res.append( begin );
                    res.append( end );
                    AlternativesPosTestGenerator.this.sendTestToProsessor( res );
                    result = new Boolean( true );

                    if( AlternativesPosTestGenerator.this.done ) return;                    
                    node.getExpression().accept( this );
                    break;
                }
            case Expression_Factor.ZERO_LIST:
                {
                    Sentence res = new Sentence();
                    res.append( begin );
                    res.append( end );
                    AlternativesPosTestGenerator.this.sendTestToProsessor( res );
                    result = new Boolean( true );
                    
                    if( AlternativesPosTestGenerator.this.done ) return;
                    node.getExpression().accept( this );
                    if( AlternativesPosTestGenerator.this.done ) return;
                    
                    Sentence res_first 
                        = AlternativesPosTestGenerator.this.generate_first( node.getExpression() );
                    if( res_first != null )
                    {
                        Sentence b = begin.makeClone();
                        Sentence e = end.makeClone();
                        begin.append( res_first );
                        node.getExpression().accept( this );
                        begin = b;
                        if( AlternativesPosTestGenerator.this.done ) return;
                        res_first.append(end);
                        end = res_first;
                        node.getExpression().accept( this );
                        end = e;
                    }
                    break;
                }
            case Expression_Factor.ONE_LIST:
                {
                    node.getExpression().accept( this );
                    if( AlternativesPosTestGenerator.this.done ) return;
                    
                    Sentence res_first 
                        = AlternativesPosTestGenerator.this.generate_first( node.getExpression() );
                    if( res_first != null )
                    {
                        Sentence b = begin.makeClone();
                        Sentence e = end.makeClone();
                        begin.append( res_first );
                        node.getExpression().accept( this );
                        begin = b;
                        if( AlternativesPosTestGenerator.this.done ) return;
                        res_first.append(end);
                        end = res_first;
                        node.getExpression().accept( this );
                        end = e;
                    }
                    break;
                }
            case Expression_Factor.DEFAULT:
                {
                    node.getExpression().accept( this );
                    break;
                }
            case Expression_Factor.TRANSPOSITION:
                {
                    Sentence res = new Sentence();
                    res.append( begin );
                    res.append( end );
                    AlternativesPosTestGenerator.this.sendTestToProsessor( res );
                    result = new Boolean( true );
                    if( AlternativesPosTestGenerator.this.done ) return;

                    node.getExpression().accept( this );
                    if( AlternativesPosTestGenerator.this.done ) return;

                    Expression expr = node.getExpression();

                    if( expr.sizeTermList() > 1 )
                    {
                        for( int i = 0; i < expr.sizeTermList(); i++ )
                        {
                            Term term_i = expr.getTermList(i);
                            boolean first_found = false;
                            
                            for( int j = 0
                               ;    j < expr.sizeTermList() 
                                 && !first_found
                               ; j++
                               )
                            {
                                if( i != j )
                                {
                                    Term term_j = expr.getTermList(j);

                                    Sentence res_first 
                                    =  AlternativesPosTestGenerator
                                      .this
                                      .generate_first( term_j );

                                    if( res_first != null )
                                    {
                                        first_found = true;

                                        Sentence b = begin.makeClone();
                                        Sentence e = end.makeClone();
    
                                        begin.append( res_first );
                                        term_i.accept( this );
                                        begin = b;
                                        if( AlternativesPosTestGenerator.this.done ) return;
                                        Sentence ee = res_first.makeClone();
                                        ee.append(end);
                                        end = ee;
                                        term_i.accept( this );
                                        end = e;
                                        if( AlternativesPosTestGenerator.this.done ) return;
                                    }
                                }
                            }
                        }
                    }
                    
                    break;
                }
            default:
                throw new IllegalArgumentException
                    ( String.valueOf( node.getKind() ) );
            }
        }

        public void visitID( ID node ) {}

        public Sentence setBegin( Sentence b )
        {
            Sentence s = begin;
            begin = b;
            return s;
        }

        public Sentence setEnd( Sentence e )
        {
            Sentence s = end;
            end = e;
            return s;
        }

        public void clear()
        {
            begin = new Sentence();
            end = new Sentence();
            result = new Boolean( false );
        }

        private Boolean result = new Boolean( false );
        private Sentence begin = new Sentence();
        private Sentence end = new Sentence();
        private NonTerminal father;
        private  int max_depth;
        private  int depth;
        private  int level;
    }

//==============================================================================

    private class ChainGenerator implements NonTerminalRealizer
    {
        public Vector/*Rule*/ setChain( Vector/*Rule*/ p_chain )
        {
            Vector/*Rule*/ c = chain;
            chain = p_chain;
            return c;
        }
        
        public int setIndex( int p_index )
        {
            int i = index;
            index = p_index;
            return i;
        }

        public NonTerminal setFather( NonTerminal p_father )
        {
            NonTerminal f = father;
            father = p_father;
            return f;
        }
    
        public Object getResult() { return result; }
        
        public void visitNode( Node node ) {}

        public void visitGrammar( Grammar node ) {}

        public void visitRule( Rule node )
        {
            result = new Boolean( false );
            if ( begin != null && end != null )
            {
                Bound beg = new Bound( node );
                begin.append( beg );
                Sentence res = new Sentence( beg.getEnd() );
                res.append( end );
                end = res;
            }
            node.getExpression().accept( this );
        }

        public void visitExpression( Expression node )
        {
            for( int i = 0; i < node.sizeTermList(); i++ )
            {
                /*
                 *     ,  
                 *   
                 */
                result = new Boolean( false );

                node.getTermList( i ).accept( this );
                if( result.booleanValue() ) return;
            }
        }

        /**
         *     :
         * -      null;
         * -  (   ,   ).
         */
        public void visitTerm( Term node )
        {
            Sentence b;
            Sentence e;

            if( begin != null && end != null )
            {
                b = begin.makeClone();
                e = end.makeClone();
            } else {
                return;
            }
            
            // find Name_Factor corresponding to the current chain element
            for( int i = 0; i < node.sizeFactorList(); i++ )
            {
                Factor factor = node.getFactorList( i );
                if ( factor instanceof Name_Factor 
                   && ((Name_Factor)factor).getDefinition().equals( 
                       (Rule)chain.elementAt( index ) 
                                                                  ) 
                   )
                {
                    // found
                    generateBeginAndEnd( node, i );
                    return;
                }
            }

            // not found. continue search in Expression_Factors
            for( int i = 0; i < node.sizeFactorList(); i++ )
            {
                if( node.getFactorList( i ) instanceof Expression_Factor )
                {
                    generateBeginAndEnd( node, i );
                    // if found stop searching
                    if( result.booleanValue() ) return;

                    // restore context
                    begin = b;
                    end = e;
                }
            }

            // unreachable statement - chain element must be found
        }

        public void visitFactor( Factor node ) {}

        public void visitName_Factor( Name_Factor node )
        {
            if( node.getDefinition().equals( (Rule)chain.elementAt( index ) ) )
            {
                /*
                 *      , 
                 *    .
                 */
                result = (Boolean)(AlternativesPosTestGenerator
                                  .this
                                  .generate_chain( chain
                                                 , index + 1
                                                 , new NonTerminal( father
                                                                  , node.getDefinition()
                                                                  )
                                                     )
                                  );
            }
         /* 
          * else {
          *           .
          * }
          */
        }

        public void visitLexeme_Factor( Lexeme_Factor node )
        {
            result = new Boolean( false );
        }

        public void visitExpression_Factor( Expression_Factor node )
        {
            node.getExpression().accept( this );
        }

        public void visitID( ID node ) {}
        
        /**
         *   (/).
         */
        private boolean generateContext( Factor factor, Sentence res )
        {
            Sentence res_first = AlternativesPosTestGenerator.this.generate_first( factor );
            if( res_first == null )
            { 
                return false;
            } else {
                res.append( res_first );
                return true;
            }    
        }

        private void generateBeginAndEnd( Term node, int number )
        {
            Sentence res = new Sentence();
            
            Factor factor = node.getFactorList( number );
            
            if ( begin != null && end != null ){
                res = begin.makeClone();
                for( int j = 0; j < number; j++ )
                {                    
                    Factor fact = node.getFactorList( j );
                    if( !generateContext( fact, res ) )
                    {
                        begin = null;
                        break;
                    }
                }
                if ( begin != null )
                {
                    begin = res;
                }
                res = new Sentence();
                for( int j = number + 1; j < node.sizeFactorList(); j++ )
                {                    
                    Factor fact = node.getFactorList( j );
                    if ( !generateContext( fact, res ) )
                    {
                        end = null;
                        break;
                    }
                }
                if ( end != null )
                {
                    res.append(end);
                    end = res;
                }
            }
            factor.accept( this );        
        }

        public Sentence getBegin() { return begin; }

        public Sentence getEnd() { return end; }

        public void clear()
        {
            begin = new Sentence();
            end = new Sentence();
            result = new Boolean( false );
        }

        private Sentence begin = new Sentence();
        private Sentence end = new Sentence();

        private Boolean result = new Boolean( false );
        private NonTerminal father;

        private Vector/*Rule*/ chain;
        private int index;
    }

    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return version;
    }
}
