/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.testgen.auxil;

//==============================================================================

import java.io.File;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Hashtable;

import com.unitesk.atp.tree.Node;
import com.unitesk.atp.text.location.Position;
import com.unitesk.atp.bnf.BNFModel.*;
import com.unitesk.atp.bnf.auxil.*;
import com.unitesk.atp.bnf.messages.*;
import com.unitesk.atp.messages.*;


//==============================================================================

public class FirstRealizationGenerator
{
    private MyFirst my_first_generator = new MyFirst();
    private FRHash first_realization_table;
    private final MessageBox mbox;
    private final LexemeTable lexeme_table;
    private final FirstGenerator first_generator;

    public FirstRealizationGenerator( Grammar grammar
                                    , String file_name
                                    , int first_depth
                                    , SentenceMarker marker
                                    , LexemeTable p_lexeme_table
                                    , MessageBox p_mbox
                                    )
    {
        mbox = p_mbox;
        lexeme_table = p_lexeme_table;
        first_generator = new FirstGenerator( marker, first_depth, lexeme_table );
        first_realization_table = new FRHash( grammar, file_name );
    }

    public Sentence generate_first( Node node )
    {
        if( node instanceof Name_Factor )
        {
            return first_realization_table.get( (Name_Factor)node );
        } else {
            node.accept( my_first_generator );
            return my_first_generator.getResult();
        }
    }

    //==========================================================================

    private class FRHash
    {
        private Hashtable/*Rule/Sentence*/ first_realization = new Hashtable();

        FRHash( Grammar grammar, String file_name )
        {
            boolean file_was_read = false;
            if( file_name != null )
            {
                File file = new File( file_name );
                if ( file.exists() )
                {
                    file_was_read = readFile( grammar, file );
                }
            }

            if( grammar.sizeOptRuleList() == first_realization.size() )
            {
                return;
            }
            mbox.info( "Constructing first realizations for "
                     + ( file_was_read
                       ? "the other"
                       : "all"
                       )
                     + " rules"
                     );
            boolean was_point = false;
            try
            {
                for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
                {
                    Rule rule = grammar.getOptRuleList( i );
                    String name = rule.getName().getValue();
                    if ( !(first_realization.containsKey( name )) )
                    {
                        Sentence res_first = getRealization( rule );
                        if ( res_first == null )
                        {
                            if( was_point )
                            {
                                System.err.println();
                                was_point = false;
                            }
                            mbox.error
                                ( new NoRealizationMessage
                                    ( rule.getName().getValue()
                                    , first_generator.getMaxDepth()
                                    )
                                );
                        } else {
                            first_realization.put( name, res_first );
                        }
                        System.err.print("."); // ???
                        was_point = true;
                    }
                }
            } finally {
                if( was_point )
                {
                    System.err.println();
                    was_point = false;
                }
            }
            if ( file_name != null )
            {
                print( grammar, file_name );
            }
        }

        private Sentence getRealization( Rule rule )
        {
            Sentence res_first = null;
            int first_max_depth = first_generator.getMaxDepth();
            int f_depth = 1;
            first_generator.setDepth( f_depth );

            while( res_first == null && f_depth < first_max_depth )
            {
                NonTerminal f
                    = first_generator.setFather( null );
                rule.accept( first_generator );
                first_generator.setFather( f );
                res_first = (Sentence)first_generator.getResult();
                f_depth++;
                first_generator.setDepth( f_depth );
            }
            return res_first;
        }

        private boolean readFile ( Grammar grammar, File file )
        {
            mbox.info( "Reading first realizations from file: " + file.getPath() );

            BufferedReader in;
            try
            {
                in = new BufferedReader( new FileReader( file ) );
            }
            catch( FileNotFoundException e )
            {
                mbox.warning( new FileNotFoundMessage( file ) );
                return false;
            }

            int line_num = 1;
            try
            {
                for( String str = null; (str = in.readLine()) != null; line_num++ )
                {
                    String s = str.trim();
                    if( s.length() > 0)
                    {
                        String name;
                        String realization;
                        int i = s.indexOf(' ');
                        if ( i == -1 )
                        {
                            name = s;
                            realization = "";
                        } else
                        {
                            name = s.substring( 0 , i );
                            realization = s.substring( i + 1, s.length() );
                        }
                        if ( grammar.getRule( name ) == null )
                        {
                            mbox.warning
                                ( new SyntaxErrorMessage( file
                                                        , new Position( line_num, 0 )
                                                        , "No such rule: '" + name + "'"
                                                        )
                                );
                        } else {
                            if( first_realization.containsKey( name ) )
                            {
                                mbox.warning
                                    ( new SyntaxErrorMessage( file
                                                            , new Position( line_num, 0 )
                                                            , "First realization for rule '"
                                                            + name + "' already defined (ignored)"
                                                            )
                                    );
                            } else {
                                first_realization.put
                                    ( name
                                    , new Sentence( new Terminal( realization.trim() ) )
                                    );
                            }
                        }
                    }
                }
                in.close();
            } catch( IOException e )
            {
                mbox.warning( new IOErrorMessage( file ) );
            }
            return true;
        }

        private void print( Grammar grammar, String file_name )
        {
            mbox.info( "Writing first realizations to file: " + file_name );

            PrintWriter out;
            try
            {
                out = new PrintWriter( new FileOutputStream( file_name ) );
            }
            catch( FileNotFoundException e )
            {
                mbox.warning( new FileNotFoundMessage( new File( file_name ) ) );
                return;
            }
            for( int i = 0; i < grammar.sizeOptRuleList(); i++ )
            {
                Rule rule = grammar.getOptRuleList( i );
                String name = rule.getName().getValue();
                if ( first_realization.containsKey( name ) )
                {
                    out.print( rule.getName().getValue() );
                    out.println( "    " + ((Sentence)first_realization.get( name )).toString() );
                }
            }
            out.close();
        }

        Sentence get( Name_Factor node )
        {
            String name = (((Name_Factor)node).getDefinition()).getName().getValue();
            if( first_realization.containsKey( name ) )
            {
                return (Sentence)first_realization.get( name );
            } else {
                return null;
            }
        }
    }

    private class MyFirst implements BNF_Visitor
    {
        public Sentence getResult() { return  result; }

        public void visitNode( Node node ){}
        public void visitGrammar( Grammar node ){}
        public void visitRule( Rule node ){}

        public void visitExpression( Expression node )
        {
            result = null;
            for( int i = 0; i < node.sizeTermList(); i++ )
            {
                node.getTermList( i ).accept( this );
                if( result != null ) return;
            }
        }

        public void visitTerm( Term node )
        {
            Sentence res = new Sentence();
            for( int i = 0; i < node.sizeFactorList(); i++ )
            {
                Factor factor = node.getFactorList( i );
                factor.accept( this );
                if( result == null ) return;
                res.append( result );
            }
            result = res;
        }

        public void visitFactor( Factor node ) {}

        public void visitName_Factor( Name_Factor node )
        {
            result = FirstRealizationGenerator.this.first_realization_table.get( node );
        }

        public void visitLexeme_Factor( Lexeme_Factor node )
        {
            result = new Sentence
                ( new Terminal( FirstRealizationGenerator.this.lexeme_table.getLexeme(node.getLexeme(), 0), node ) );
        }

        public void visitExpression_Factor( Expression_Factor node )
        {
            switch( node.getKind() )
            {
            case Expression_Factor.OPTIONAL:
            case Expression_Factor.ZERO_LIST:
            case Expression_Factor.TRANSPOSITION:
                result = new Sentence();
                break;
            case Expression_Factor.ONE_LIST:
            case Expression_Factor.DEFAULT:
                node.getExpression().accept( this );
                break;
            default:
                throw new IllegalArgumentException
                    ( String.valueOf( node.getKind() ) );
            }
        }

        public void visitID( ID node ) {}

        private Sentence result;
    }

}
