/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.testgen.backend;

//==============================================================================

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;

import com.unitesk.atp.bnf.actions.testgen.auxil.Sentence;
import com.unitesk.atp.bnf.actions.testgen.auxil.SentenceList;
import com.unitesk.atp.bnf.actions.testgen.auxil.Terminal;
import com.unitesk.atp.bnf.actions.testgen.auxil.Word;
import com.unitesk.atp.bnf.auxil.LexemeTable;
import com.unitesk.atp.bnf.messages.*;




public class XMLPrinter extends TestCaseProcessor
{
    private static String version = "1.1.2";

    private boolean was_init = false;

    private PrintWriter xml_out;
    private String output_dir;
    private String file_prename;
    private String file_postname;

    private boolean show_derivation_tree = true;
    private boolean xml_only = false;
    private String target_id = null;

    public void process( Object tests, String kind )
    {
        if ( !was_init )
        {
            init( kind );
            was_init = true;
        }

        if ( tests instanceof SentenceList )
        {
            printList( (SentenceList)tests, kind );
        } else if ( tests instanceof Sentence ) {
            printSentence( (Sentence)tests, kind );
        } else {
            throw new ClassCastException( tests.getClass().getName() );
        }
    }

    private void printList( SentenceList list, String kind )
    {
        int size = list.size();
        for( int i = 0; i < size; i++ )
        {
            printSentence( list.get( i ), kind );
        }
    }

    private String getDirName( int n )
    {
        StringBuffer res = new StringBuffer();
        for( char  d: Integer.toString(n).toCharArray() )
        {
            res.append( d );
            res.append( '/' );
        }
        return res.toString();
    }

    private void printSentence( Sentence s, String kind )
    {
        String file_name = output_dir + "/" + getDirName(n) + kind + "_" + file_prename + n + file_postname;
        n++;

        String test_string
            = separator_before + s.toString( separator_between ) + separator_after;

        xml_out.print( "<test kind=\"" + kind + "\""
                     + ( xml_only
                       ? ""
                       : " href=\"" + file_name + "\""
                       )
                     + ">"
                     );
        if( !show_derivation_tree )
        {
            xml_out.print( txt2xml.convert( test_string ) );
        } else {
            boolean was_term = false;
            xml_out.print( xml_separator_before );
            for( int i = 0; i < s.size(); i++ )
            {
                Word w = s.get( i );
                if( w instanceof Terminal )
                {
                    if( was_term ) xml_out.print( xml_separator_between );
                    xml_out.print( txt2xml.convert( ((Terminal)w).get() ) );
                    was_term = true;
                } else {
                    w.print( xml_out );
                }
            }
            xml_out.print( xml_separator_after );
        }
        xml_out.println( "</test>" );
        xml_out.flush();

        if( !xml_only )
        {
            PrintWriter out = openWriter( file_name );
            out.print( test_string );
            out.close();
        }
    }

    private PrintWriter openWriter( String file_name )
    {
        try
        {
            File file = new File( file_name );
            file.getCanonicalFile().getParentFile().mkdirs();
            return new PrintWriter( new FileOutputStream( file ) );
        } catch( FileNotFoundException e ) {
            throw new TestCaseProcessorException( e );
        } catch( IOException e ) {
            throw new TestCaseProcessorException( e );
        }
    }

    private int n = 0; //  

    private String separator_before;
    private String separator_between;
    private String separator_after;
    
    private String xml_separator_before;
    private String xml_separator_between;
    private String xml_separator_after;

    private void init( String kind )
    {
        n = 0;

        mbox.info( getName() + ": Initialization" );
        StringBuffer res = new StringBuffer();
        if(  LexemeTable.getValue( getProperty( "test.separator.before", "" ) 
                                 + "\""
                                 , 0
                                 , res
                                 )
          != -1
          )
        {
            separator_before = res.toString();
            xml_separator_before = txt2xml.convert( separator_before );
        } else {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "test.separator.before"
                                                , "Incorrect separator"
                                                )
                      );
        }
        res.setLength( 0 );
        if(  LexemeTable.getValue( getProperty( "test.separator.between", " " ) 
                                + "\""
                                , 0
                                , res
                                )
          != -1
          )
        {
            separator_between = res.toString();
            xml_separator_between = txt2xml.convert( separator_between );
        } else {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "test.separator.between"
                                                , "Incorrect separator"
                                                )
                      );
        }
        res.setLength( 0 );
        if(  LexemeTable.getValue( getProperty( "test.separator.after", "" ) 
                                 + "\""
                                 , 0
                                 , res
                                 )
          != -1
          )
        {
            separator_after = res.toString();
            xml_separator_after = txt2xml.convert( separator_after );
        } else {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "test.separator.after"
                                                , "Incorrect separator"
                                                )
                      );
        }

        show_derivation_tree
            = Boolean.valueOf( getProperty( "derivation.tree", "true" ) ).booleanValue();
        xml_only = Boolean.valueOf( getProperty( "only" ) ).booleanValue();
        target_id = getProperty( "target" );

        file_prename = getProperty( "file.prename", "" );
        file_postname = getProperty( "file.postname", ".test" );

        output_dir = getProperty( "output.dir", "." );
        if( !output_dir.endsWith( File.separator ) )
        {
            output_dir += File.separator;
        }

        mbox.status();
        mbox.info( getName() + ": Generating tests to " + output_dir );

        xml_out = openWriter( output_dir + kind + "_tests.xml" );

        xml_out.println( "<test_suite"
                       + ( target_id == null
                         ? ""
                         : " target=\"" + target_id + "\""
                         )
                       + ">"
                       );
    }

    public void close()
    {
        if ( xml_out != null )
        {
            xml_out.println( "</test_suite>" );
            xml_out.close();
            mbox.info( getName() + ": Number of all printed tests = " + n );
        }
        was_init = false;
    }

    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return version;
    }

    private static TextToXMLConverter txt2xml = new TextToXMLConverter();

    public static class TextToXMLConverter
    {
        private static final org.xmlpull.v1.XmlSerializer to_xml;
        static {
            try {
                to_xml = new org.xmlpull.mxp1.MXParserFactory().newSerializer();
            } catch( org.xmlpull.v1.XmlPullParserException e ) {
                throw new RuntimeException( "Can not initialize XmlSerializer" );
            }
        }
        public String convert( String arg )
        {
            java.io.StringWriter out = new java.io.StringWriter();
            try {
                to_xml.setOutput( out );
                to_xml.text( arg );
            } catch( java.io.IOException e ) {
                throw new RuntimeException( "Can not apply XmlSerializer" );
            }
            return out.toString();
        }
    }
}
