/*
 * Copyright 2001-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
//==============================================================================

package com.unitesk.atp.bnf.actions.testgen.backend;

//==============================================================================

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.FileOutputStream;
import java.io.PrintWriter;

import java.util.Vector;

import com.unitesk.atp.bnf.actions.testgen.auxil.Sentence;
import com.unitesk.atp.bnf.actions.testgen.auxil.SentenceList;
import com.unitesk.atp.bnf.actions.testgen.auxil.Terminal;
import com.unitesk.atp.bnf.actions.testgen.auxil.Word;
import com.unitesk.atp.bnf.auxil.LexemeTable;
import com.unitesk.atp.bnf.messages.*;



public class Executor extends TestCaseProcessor
{
    private static String version = "1.1.2";

    private boolean was_init = false;

    private String output_dir;
    private String command;
    private String file_prename;
    private String file_postname;
    private boolean delete_if_success = false;

    private boolean show_derivation_tree = true;
    private boolean with_xml = true;
    private String target_id = null;

    public void process( Object tests, String kind )
    {
        if ( !was_init )
        {
            init( kind );
            was_init = true;
        }

        if ( tests instanceof SentenceList )
        {
            printList( (SentenceList)tests, kind );
        } else if ( tests instanceof Sentence ) {
            printSentence( (Sentence)tests, kind );
        } else {
            throw new ClassCastException( tests.getClass().getName() );
        }
    }
    
    private void printList( SentenceList list, String kind )
    {
        int size = list.size();
        for( int i = 0; i < size; i++ )
        {
            printSentence( list.get( i ), kind );
        }
    }

    private void printSentence( Sentence s, String kind )
    {
        String file_name = output_dir + kind + "_" + file_prename + n + file_postname;
        String xml_file_name = output_dir + kind + "_" + file_prename + n + file_postname + ".xml";

        n++;

        String test_string
            = separator_before + s.toString( separator_between ) + separator_after;

        if( with_xml )
        {
            PrintWriter xml_out;
            try
            {
                xml_out = new PrintWriter( new FileOutputStream( xml_file_name ) );
            }
            catch( FileNotFoundException e )
            {
                throw new TestCaseProcessorException( e );
            }
            xml_out.print( "<test kind=\"" + kind + "\" href=\"" + file_name + "\""
                         + ( target_id == null
                           ? ""
                           : " target=\"" + target_id + "\""
                           )
                         + ">"
                         );
            if( !show_derivation_tree )
            {
                xml_out.print( txt2xml.convert( test_string ) );
            } else {
                boolean was_term = false;
                xml_out.print( xml_separator_before );
                for( int i = 0; i < s.size(); i++ )
                {
                    Word w = s.get( i );
                    if( w instanceof Terminal )
                    {
                        if( was_term ) xml_out.print( xml_separator_between );
                        xml_out.print( txt2xml.convert( ((Terminal)w).get() ) );
                        was_term = true;
                    } else {
                        w.print( xml_out );
                    }
                }
                xml_out.print( xml_separator_after );
            }
            xml_out.println( "</test>" );
            xml_out.close();
        }

        PrintWriter out;
        try
        {
            out = new PrintWriter( new FileOutputStream( file_name ) );
        }
        catch( FileNotFoundException e )
        {
            throw new TestCaseProcessorException( e );
        }
        out.print( test_string );
        out.close();

        Vector cmd = new Vector();
        cmd.add( command );
        cmd.add( file_name );
        if( with_xml )
        {
            cmd.add( xml_file_name );
        }
        String[] actual_command = (String[])cmd.toArray( new String[0] );
        boolean verdict = false;
        try
        {
            Process p = Runtime.getRuntime().exec( actual_command );
            p.waitFor();
            verdict = ( p.exitValue() == 0 );
        } catch ( IOException e ) {
            throw new TestCaseProcessorException( e );
        } catch ( InterruptedException e ) {
            throw new TestCaseProcessorException( e );
        }

        if( verdict )
        {
            n_success++;
            if( delete_if_success )
            {
                new File( file_name ).delete();
                if( with_xml ) { new File( xml_file_name ).delete(); }
            }
        }
    }

    private int n = 0; // ����� �����
    private int n_success = 0; // ���������� �������� ��������

    private String separator_before;
    private String separator_between;
    private String separator_after;
    
    private String xml_separator_before;
    private String xml_separator_between;
    private String xml_separator_after;

    private void init( String kind )
    {
        n = 0;
        n_success = 0;

        mbox.info( getName() + ": Initialization" );
        StringBuffer res = new StringBuffer();
        if(  LexemeTable.getValue( getProperty( "test.separator.before", "" )
                                 + "\""
                                 , 0
                                 , res
                                 ) 
          != -1
          )
        {
            separator_before = res.toString();
            xml_separator_before = txt2xml.convert( separator_before );
        } else {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "test.separator.before"
                                                , "Incorrect separator"
                                                )
                      );
        }
        res.setLength( 0 );
        if(  LexemeTable.getValue( getProperty( "test.separator.between", " " ) 
                                 + "\""
                                 , 0
                                 , res
                                 )
          != -1
          )
        {
            separator_between = res.toString();
            xml_separator_between = txt2xml.convert( separator_between );
        } else {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "test.separator.between"
                                                , "Incorrect separator"
                                                )
                      );
        }
        res.setLength( 0 );
        if(  LexemeTable.getValue( getProperty( "test.separator.after", "" ) 
                                 + "\""
                                 , 0
                                 , res
                                 )
          != -1
          )
        {
            separator_after = res.toString();
            xml_separator_after = txt2xml.convert( separator_after );
        } else {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "test.separator.after"
                                                , "Incorrect separator"
                                                )
                      );
        }

        command = getProperty( "command" );
        if( command == null )
        {
            mbox.error( new PropertyErrorMessage( getName()
                                                , "command"
                                                , "Executor command not found"
                                                )
                      );
        }

        show_derivation_tree
            = Boolean.valueOf( getProperty( "derivation.tree", "true" ) ).booleanValue();
        with_xml = Boolean.valueOf( getProperty( "with.xml", "true" ) ).booleanValue();
        target_id = getProperty( "target" );

        delete_if_success = Boolean.valueOf( getProperty( "delete", "false" ) ).booleanValue();
        file_prename = getProperty( "file.prename", "" );
        file_postname = getProperty( "file.postname", ".test" );

        output_dir = getProperty( "output.dir", "." );
        if( !output_dir.endsWith( File.separator ) )
        {
            output_dir += File.separator;
        }

        mbox.status();
        mbox.info( getName() + ": Generating tests to " + output_dir );
    }

    public void close()
    {
        mbox.info( getName() + ": Number of all executed tests = " + n );
        mbox.info( getName() + ": Number of successfully executed tests = " + n_success );
        mbox.info( getName() + ": Number of faults = " + (n - n_success) );
        was_init = false;
    }

    private static XMLPrinter.TextToXMLConverter txt2xml = new XMLPrinter.TextToXMLConverter();

    /**
     * {@inheritDoc}
     */
    public String getVersion()
    {
        return version;
    }
}
